# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools

import flask
from oslo_log import log
from oslo_policy import policy as common_policy
from oslo_utils import strutils

from keystone.common import authorization
from keystone.common import context
from keystone.common import policies
from keystone.common import provider_api
from keystone.common import utils
import keystone.conf
from keystone import exception
from keystone.i18n import _

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
PROVIDER_APIS = provider_api.ProviderAPIs


_POSSIBLE_TARGET_ACTIONS = frozenset(
    [
        rule.name
        for rule in policies.list_rules()
        if not rule.deprecated_for_removal
    ]
)
_ENFORCEMENT_CHECK_ATTR = 'keystone:RBAC:enforcement_called'


class RBACEnforcer:
    """Enforce RBAC on API calls."""

    __shared_state__: dict = {}
    __ENFORCER = None
    ACTION_STORE_ATTR = 'keystone:RBAC:action_name'
    # FOR TESTS ONLY
    suppress_deprecation_warnings = False

    def __init__(self):
        # NOTE(morgan): All Enforcer Instances use the same shared state;
        # BORG pattern.
        self.__dict__ = self.__shared_state__

    def _check_deprecated_rule(self, action):
        def _name_is_changing(rule):
            deprecated_rule = rule.deprecated_rule
            return (
                deprecated_rule
                and deprecated_rule.name != rule.name
                and deprecated_rule.name in self._enforcer.file_rules
            )

        def _check_str_is_changing(rule):
            deprecated_rule = rule.deprecated_rule
            return (
                deprecated_rule
                and deprecated_rule.check_str != rule.check_str
                and rule.name not in self._enforcer.file_rules
            )

        def _is_deprecated_for_removal(rule):
            return (
                rule.deprecated_for_removal
                and rule.name in self._enforcer.file_rules
            )

        def _emit_warning():
            if not self._enforcer._warning_emitted:
                LOG.warning(
                    "Deprecated policy rules found. Use "
                    "oslopolicy-policy-generator and "
                    "oslopolicy-policy-upgrade to detect and resolve "
                    "deprecated policies in your configuration."
                )
                self._enforcer._warning_emitted = True

        registered_rule = self._enforcer.registered_rules.get(action)

        if not registered_rule:
            return
        if (
            _name_is_changing(registered_rule)
            or _check_str_is_changing(registered_rule)
            or _is_deprecated_for_removal(registered_rule)
        ):
            _emit_warning()

    def _enforce(self, credentials, action, target, do_raise=True):
        """Verify that the action is valid on the target in this context.

        This method is for cases that exceed the base enforcer
        functionality (notably for compatibility with `@protected` style
        decorators.

        :param credentials: user credentials
        :param action: string representing the action to be checked, which
                       should be colon separated for clarity.
        :param target: dictionary representing the object of the action for
                       object creation this should be a dictionary
                       representing the location of the object e.g.
                       {'project_id': object.project_id}
        :raises keystone.exception.Forbidden: If verification fails.

        Actions should be colon separated for clarity. For example:

        * identity:list_users
        """
        # Add the exception arguments if asked to do a raise
        extra = {}
        if do_raise:
            extra.update(
                exc=exception.ForbiddenAction, action=action, do_raise=do_raise
            )

        try:
            result = self._enforcer.enforce(
                rule=action, target=target, creds=credentials, **extra
            )
            self._check_deprecated_rule(action)
            return result
        except common_policy.InvalidScope:
            raise exception.ForbiddenAction(action=action)

    def _reset(self):
        # NOTE(morgan): Used for TEST purposes only.
        self.__ENFORCER = None

    @property
    def _enforcer(self):
        # The raw oslo-policy enforcer object
        if self.__ENFORCER is None:
            self.__ENFORCER = common_policy.Enforcer(CONF)
            # NOTE(cmurphy) when running in the keystone server, suppress
            # deprecation warnings for individual policy rules. Instead, we log
            # a single notification at enforcement time indicating the
            # oslo.policy tools the operator can use to detect and resolve
            # deprecated policies. If there is no request context here, that
            # means external tooling such as the oslo.policy tools are running
            # this code, in which case we do want the full deprecation warnings
            # emitted for individual polcy rules.
            if flask.has_request_context():
                self.__ENFORCER.suppress_deprecation_warnings = True
            # NOTE(cmurphy) Tests may explicitly disable these warnings to
            # prevent an explosion of test logs
            if self.suppress_deprecation_warnings:
                self.__ENFORCER.suppress_deprecation_warnings = True
            self.register_rules(self.__ENFORCER)
            self.__ENFORCER._warning_emitted = False
        return self.__ENFORCER

    @staticmethod
    def _extract_filter_values(filters):
        """Extract filter data from query params for RBAC enforcement."""
        filters = filters or []
        target = {
            i: flask.request.args[i]
            for i in filters
            if i in flask.request.args
        }
        if target:
            if LOG.logger.getEffectiveLevel() <= log.DEBUG:
                LOG.debug(
                    'RBAC: Adding query filter params (%s)',
                    ', '.join([f'{k}={v}' for k, v in target.items()]),
                )
        return target

    @staticmethod
    def _extract_member_target_data(member_target_type, member_target):
        """Build some useful target data.

        :param member_target_type: what type of target, e.g. 'user'
        :type member_target_type: str or None
        :param member_target: reference of the target data
        :type member_target: dict or None
        :returns: constructed target dict or empty dict
        :rtype: dict
        """
        ret_dict = {}
        if (member_target is not None and member_target_type is None) or (
            member_target is None and member_target_type is not None
        ):
            LOG.warning(
                'RBAC: Unknown target type or target reference. '
                'Rejecting as unauthorized. '
                '(member_target_type=%(target_type)r, '
                'member_target=%(target_ref)r)',
                {
                    'target_type': member_target_type,
                    'target_ref': member_target,
                },
            )
            # Fast exit.
            return ret_dict

        if member_target is not None and member_target_type is not None:
            ret_dict['target'] = {member_target_type: member_target}
        else:
            # Try and do some magic loading based upon the resource we've
            # matched in our route. This is mostly so we can have a level of
            # automatic pulling in the resource; strictly for some added
            # DRY capabilities. In an ideal world the target is always passed
            # in explicitly.
            if flask.request.endpoint:
                # This only works for cases of Flask-RESTful, or carefully
                # crafted endpoints that live on a class. Ultimately, there
                # should be more protection against something wonky
                # here.
                resource = flask.current_app.view_functions[
                    flask.request.endpoint
                ].view_class
                try:
                    member_name = getattr(resource, 'member_key', None)
                except ValueError:
                    # NOTE(morgan): In the case that the ResourceBase keystone
                    # class is used, we raise a value error when member_key
                    # has not been set on the class. This is perfectly
                    # normal and acceptable. Set member_name to None as though
                    # it wasn't set.
                    member_name = None
                func = getattr(resource, 'get_member_from_driver', None)
                if member_name is not None and callable(func):
                    key = f'{member_name}_id'
                    if key in (flask.request.view_args or {}):
                        # NOTE(morgan): For most correct setup, instantiate the
                        # view_class. There is no current support for passing
                        # extra args to the constructor of the view_class like
                        # .as_view() method would actually do. In this case
                        # perform a simple instantiation to represent the
                        # `self` pass to the unbound method.
                        #
                        # TODO(morgan): add (future) support for passing class
                        # instantiation args.
                        ret_dict['target'] = {
                            member_name: func(flask.request.view_args[key])
                        }
        return ret_dict

    @staticmethod
    def _extract_policy_check_credentials():
        # Pull out the auth context
        return flask.request.environ.get(authorization.AUTH_CONTEXT_ENV, {})

    @classmethod
    def _extract_subject_token_target_data(cls):
        ret_dict = {}
        window_seconds = 0
        # NOTE(morgan): Populate the subject token data into
        # the policy dict at "target.token". In all liklyhood
        # it is un-interesting to populate this data outside
        # of the auth paths.
        target = 'token'
        subject_token = flask.request.headers.get('X-Subject-Token')
        access_rules_support = flask.request.headers.get(
            authorization.ACCESS_RULES_HEADER
        )
        if subject_token is not None:
            allow_expired = strutils.bool_from_string(
                flask.request.args.get('allow_expired', False), default=False
            )
            if allow_expired:
                window_seconds = CONF.token.allow_expired_window
            token = PROVIDER_APIS.token_provider_api.validate_token(
                subject_token,
                window_seconds=window_seconds,
                access_rules_support=access_rules_support,
            )
            # TODO(morgan): Expand extracted data from the subject token.
            ret_dict[target] = {}
            ret_dict[target]['user_id'] = token.user_id
            try:
                user_domain_id = token.user['domain_id']
            except exception.UnexpectedError:
                user_domain_id = None
            if user_domain_id:
                ret_dict[target].setdefault('user', {})
                ret_dict[target]['user'].setdefault('domain', {})
                ret_dict[target]['user']['domain']['id'] = user_domain_id
        return ret_dict

    @staticmethod
    def _get_oslo_req_context():
        return flask.request.environ.get(context.REQUEST_CONTEXT_ENV, None)

    @classmethod
    def _assert_is_authenticated(cls):
        ctx = cls._get_oslo_req_context()
        if ctx is None:
            LOG.warning(
                'RBAC: Error reading the request context generated by '
                'the Auth Middleware (there is no context). Rejecting '
                'request as unauthorized.'
            )
            raise exception.Unauthorized(
                _(
                    'Internal error processing authentication and '
                    'authorization.'
                )
            )
        if not ctx.authenticated:
            raise exception.Unauthorized(
                _('auth_context did not decode anything useful')
            )

    @classmethod
    def _shared_admin_auth_token_set(cls):
        ctx = cls._get_oslo_req_context()
        return getattr(ctx, 'is_admin', False)

    @classmethod
    def enforce_call(
        cls,
        enforcer=None,
        action=None,
        target_attr=None,
        member_target_type=None,
        member_target=None,
        filters=None,
        build_target=None,
    ):
        """Enforce RBAC on the current request.

        This will do some legwork and then instantiate the Enforcer if an
        enforcer is not passed in.

        :param enforcer: A pre-instantiated Enforcer object (optional)
        :type enforcer: :class:`RBACEnforcer`
        :param action: the name of the rule/policy enforcement to be checked
                       against, e.g. `identity:get_user` (optional may be
                       replaced by decorating the method/function with
                       `policy_enforcer_action`.
        :type action: str
        :param target_attr: complete override of the target data. This will
                            replace all other generated target data meaning
                            `member_target_type` and `member_target` are
                            ignored. This will also prevent extraction of
                            data from the X-Subject-Token. The `target` dict
                            should contain a series of key-value pairs such
                            as `{'user': user_ref_dict}`.
        :type target_attr: dict
        :param member_target_type: the type of the target, e.g. 'user'. Both
                                   this and `member_target` must be passed if
                                   either is passed.
        :type member_target_type: str
        :param member_target: the (dict form) reference of the member object.
                              Both this and `member_target_type` must be passed
                              if either is passed.
        :type member_target: dict
        :param filters: A variable number of optional string filters, these are
                        used to extract values from the query params. The
                        filters are added to the request data that is passed to
                        the enforcer and may be used to determine policy
                        action. In practice these are mainly supplied in the
                        various "list" APIs and are un-used in the default
                        supplied policies.
        :type filters: iterable
        :param build_target: A function to build the target for enforcement.
                             This is explicitly done after authentication
                             in order to not leak existance data before
                             auth.
        :type build_target: function
        """
        # NOTE(morgan) everything in the policy_dict may be used by the policy
        # DSL to action on RBAC and request information/response data.
        policy_dict = {}

        # If "action" has not explicitly been overridden, see if it is set in
        # Flask.g app-context (per-request thread local) meaning the
        # @policy_enforcer_action decorator was used.
        action = action or getattr(flask.g, cls.ACTION_STORE_ATTR, None)
        if action not in _POSSIBLE_TARGET_ACTIONS:
            LOG.warning(
                'RBAC: Unknown enforcement action name `%s`. '
                'Rejecting as Forbidden, this is a programming error '
                'and a bug should be filed with as much information '
                'about the request that caused this as possible.',
                action,
            )
            # NOTE(morgan): While this is an internal error, a 500 is never
            # desirable, we have handled the case and the most appropriate
            # response here is to issue a 403 (FORBIDDEN) to any API calling
            # enforce_call with an inappropriate action/name to look up the
            # policy rule. This is simply a short-circuit as the enforcement
            # code raises a 403 on an unknown action (in keystone) by default.
            raise exception.Forbidden(
                message=_(
                    'Internal RBAC enforcement error, invalid rule (action) '
                    'name.'
                )
            )

        # Mark flask.g as "enforce_call" has been called. This should occur
        # before anything except the "is this a valid action" check, ensuring
        # all proper "after request" checks pass, showing that the API has
        # enforcement.
        setattr(flask.g, _ENFORCEMENT_CHECK_ATTR, True)

        # Assert we are actually authenticated
        cls._assert_is_authenticated()

        # Check if "is_admin", this is in support of the old "admin auth token"
        # middleware with a shared "admin" token for auth
        if cls._shared_admin_auth_token_set():
            LOG.warning('RBAC: Bypassing authorization')
            return

        # NOTE(morgan): !!! ORDER OF THESE OPERATIONS IS IMPORTANT !!!
        # The lowest priority values are set first and the highest priority
        # values are set last.

        # Populate the input attributes (view args) directly to the policy
        # dict. This is to allow the policy engine to have access to the
        # view args for substitution. This is to mirror the old @protected
        # mechanism and ensure current policy files continue to work as
        # expected.
        policy_dict.update(flask.request.view_args)

        # Get the Target Data Set.
        if target_attr is None and build_target is None:
            try:
                policy_dict.update(
                    cls._extract_member_target_data(
                        member_target_type, member_target
                    )
                )
            except exception.NotFound:
                # DEBUG LOG and bubble up the 404 error. This is expected
                # behavior. This likely should be specific in each API. This
                # should be revisited in the future and each API should make
                # the explicit "existence" checks before enforcement.
                LOG.debug(
                    'Extracting inferred target data resulted in '
                    '"NOT FOUND (404)".'
                )
                raise
            except Exception as e:  # nosec
                # NOTE(morgan): Errors should never bubble up at this point,
                # if there is an error getting the target, log it and move
                # on. Raise an explicit 403, we have failed policy checks.
                LOG.warning(
                    'Unable to extract inferred target data during '
                    'enforcement'
                )
                LOG.debug(e, exc_info=True)
                raise exception.ForbiddenAction(action=action)

            # Special Case, extract and add subject_token data.
            subj_token_target_data = cls._extract_subject_token_target_data()
            if subj_token_target_data:
                policy_dict.setdefault('target', {}).update(
                    subj_token_target_data
                )
        else:
            if target_attr and build_target:
                raise ValueError(
                    'Programming Error: A target_attr or '
                    'build_target must be provided, but not both'
                )

            policy_dict['target'] = target_attr or build_target()

        # Pull the data from the submitted json body to generate
        # appropriate input/target attributes, we take an explicit copy here
        # to ensure we're not somehow corrupting
        json_input = flask.request.get_json(force=True, silent=True) or {}
        policy_dict.update(json_input.copy())

        # Generate the filter_attr dataset.
        policy_dict.update(cls._extract_filter_values(filters))

        flattened = utils.flatten_dict(policy_dict)
        if LOG.logger.getEffectiveLevel() <= log.DEBUG:
            # LOG the Args
            args_str = ', '.join(
                [
                    f'{k}={v}'
                    for k, v in (flask.request.view_args or {}).items()
                ]
            )
            args_str = strutils.mask_password(args_str)
            LOG.debug(
                'RBAC: Authorizing `%(action)s(%(args)s)`',
                {'action': action, 'args': args_str},
            )

        ctxt = cls._get_oslo_req_context()
        # Instantiate the enforcer object if needed.
        enforcer_obj = enforcer or cls()
        enforcer_obj._enforce(
            credentials=ctxt, action=action, target=flattened
        )
        LOG.debug('RBAC: Authorization granted')

    @classmethod
    def policy_enforcer_action(cls, action):
        """Decorator to set policy enforcement action name."""
        if action not in _POSSIBLE_TARGET_ACTIONS:
            raise ValueError(
                'PROGRAMMING ERROR: Action must reference a '
                'valid Keystone policy enforcement name.'
            )

        def wrapper(f):
            @functools.wraps(f)
            def inner(*args, **kwargs):
                # Set the action in g on a known attr so we can reference it
                # later.
                setattr(flask.g, cls.ACTION_STORE_ATTR, action)
                return f(*args, **kwargs)

            return inner

        return wrapper

    @staticmethod
    def register_rules(enforcer):
        enforcer.register_defaults(policies.list_rules())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import uuid

from oslo_serialization import jsonutils

from keystone.common.policies import base
from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _SystemUserSystemAssignmentTests:
    def test_user_can_list_user_system_role_assignments(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/system/users/{}/roles'.format(user['id']),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))
            self.assertEqual(
                self.bootstrapper.member_role_id, r.json['roles'][0]['id']
            )

    def test_user_can_check_user_system_role_assignment(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.get(
                '/v3/system/users/{}/roles/{}'.format(
                    user['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_user_can_list_group_system_role_assignments(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_group(
            group['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/system/groups/{}/roles'.format(group['id']),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))
            self.assertEqual(
                self.bootstrapper.member_role_id, r.json['roles'][0]['id']
            )

    def test_user_can_check_group_system_role_assignments(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_group(
            group['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.get(
                '/v3/system/groups/{}/roles/{}'.format(
                    group['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )


class _SystemMemberAndReaderSystemAssignmentTests:
    def test_user_cannot_grant_system_assignments(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/system/users/{}/roles/{}'.format(
                    user['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_revoke_system_assignments(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.delete(
                '/v3/system/users/{}/roles/{}'.format(
                    user['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_grant_group_system_assignment(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/system/groups/{}/roles/{}'.format(
                    group['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_revoke_group_system_assignment(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_group(
            group['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.delete(
                '/v3/system/groups/{}/roles/{}'.format(
                    group['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _DomainAndProjectUserSystemAssignmentTests:
    def test_user_cannot_list_system_role_assignments(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.get(
                '/v3/system/users/{}/roles'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_check_user_system_role_assignments(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.get(
                '/v3/system/users/{}/roles/{}'.format(
                    user['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_grant_system_assignments(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/system/users/{}/roles/{}'.format(
                    user['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_revoke_system_assignments(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.delete(
                '/v3/system/users/{}/roles/{}'.format(
                    user['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_group_system_role_assignments(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_group(
            group['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.get(
                '/v3/system/groups/{}/roles'.format(group['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_check_group_system_role_assignments(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_group(
            group['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.get(
                '/v3/system/groups/{}/roles/{}'.format(
                    group['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_grant_group_system_assignments(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/system/groups/{}/roles/{}'.format(
                    group['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_revoke_group_system_assignments(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_group(
            group['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.delete(
                '/v3/system/groups/{}/roles/{}'.format(
                    group['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class SystemReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserSystemAssignmentTests,
    _SystemMemberAndReaderSystemAssignmentTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserSystemAssignmentTests,
    _SystemMemberAndReaderSystemAssignmentTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )
        self.expected = [
            # assignment of the user running the test case
            {
                'user_id': self.user_id,
                'system': 'all',
                'role_id': self.bootstrapper.member_role_id,
            }
        ]

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserSystemAssignmentTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        self.expected = []

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_grant_system_assignments(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/system/users/{}/roles/{}'.format(
                    user['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
            )

    def test_user_can_revoke_system_assignments(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.delete(
                '/v3/system/users/{}/roles/{}'.format(
                    user['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
            )

    def test_user_can_grant_group_system_assignments(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/system/groups/{}/roles/{}'.format(
                    group['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
            )

    def test_user_can_revoke_group_system_assignments(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(CONF.identity.default_domain_id)
        )

        PROVIDERS.assignment_api.create_system_grant_for_group(
            group['id'], self.bootstrapper.member_role_id
        )

        with self.test_client() as c:
            c.delete(
                '/v3/system/groups/{}/roles/{}'.format(
                    group['id'], self.bootstrapper.member_role_id
                ),
                headers=self.headers,
            )


class DomainUserTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainAndProjectUserSystemAssignmentTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_user = unit.new_user_ref(domain_id=self.domain_id)
        self.domain_user_id = PROVIDERS.identity_api.create_user(domain_user)[
            'id'
        ]
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.domain_user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.domain_user_id,
            password=domain_user['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainAndProjectUserSystemAssignmentTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        project_reader = unit.new_user_ref(domain_id=self.domain_id)
        project_reader_id = PROVIDERS.identity_api.create_user(project_reader)[
            'id'
        ]
        project = unit.new_project_ref(domain_id=self.domain_id)
        project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=project_reader_id,
            project_id=project_id,
        )

        auth = self.build_authentication_request(
            user_id=project_reader_id,
            password=project_reader['password'],
            project_id=project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainAndProjectUserSystemAssignmentTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        project_member = unit.new_user_ref(domain_id=self.domain_id)
        project_member_id = PROVIDERS.identity_api.create_user(project_member)[
            'id'
        ]
        project = unit.new_project_ref(domain_id=self.domain_id)
        project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=project_member_id,
            project_id=project_id,
        )

        auth = self.build_authentication_request(
            user_id=project_member_id,
            password=project_member['password'],
            project_id=project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainAndProjectUserSystemAssignmentTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        project_admin = unit.new_user_ref(domain_id=self.domain_id)
        project_admin_id = PROVIDERS.identity_api.create_user(project_admin)[
            'id'
        ]
        project = unit.new_project_ref(domain_id=self.domain_id)
        project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=project_admin_id,
            project_id=project_id,
        )

        auth = self.build_authentication_request(
            user_id=project_admin_id,
            password=project_admin['password'],
            project_id=project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def _override_policy(self):
        # TODO(lbragstad): Remove this once the deprecated policies in
        # keystone.common.policies.grants have been removed. This is only
        # here to make sure we test the new policies instead of the deprecated
        # ones. Oslo.policy will OR deprecated policies with new policies to
        # maintain compatibility and give operators a chance to update
        # permissions or update policies without breaking users. This will
        # cause these specific tests to fail since we're trying to correct this
        # broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:check_system_grant_for_user': base.SYSTEM_READER,
                'identity:list_system_grants_for_user': base.SYSTEM_READER,
                'identity:create_system_grant_for_user': base.SYSTEM_ADMIN,
                'identity:revoke_system_grant_for_user': base.SYSTEM_ADMIN,
            }
            f.write(jsonutils.dumps(overridden_policies))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Internal implementation of request/response validating middleware."""

import re

import jsonschema
from jsonschema import exceptions as jsonschema_exc
from oslo_utils import timeutils
from oslo_utils import uuidutils
import webob.exc

from keystone.common import utils
from keystone import exception
from keystone.i18n import _


def _soft_validate_additional_properties(
    validator, additional_properties_value, param_value, schema
):
    """Validator function.

    If there are not any properties on the param_value that are not specified
    in the schema, this will return without any effect. If there are any such
    extra properties, they will be handled as follows:

    - if the validator passed to the method is not of type "object", this
      method will return without any effect.
    - if the 'additional_properties_value' parameter is True, this method will
      return without any effect.
    - if the schema has an additionalProperties value of True, the extra
      properties on the param_value will not be touched.
    - if the schema has an additionalProperties value of False and there
      aren't patternProperties specified, the extra properties will be stripped
      from the param_value.
    - if the schema has an additionalProperties value of False and there
      are patternProperties specified, the extra properties will not be
      touched and raise validation error if pattern doesn't match.
    """
    if not (
        validator.is_type(param_value, "object") or additional_properties_value
    ):
        return

    properties = schema.get("properties", {})
    patterns = "|".join(schema.get("patternProperties", {}))
    extra_properties = set()
    for prop in param_value:
        if prop not in properties:
            if patterns:
                if not re.search(patterns, prop):
                    extra_properties.add(prop)
            else:
                extra_properties.add(prop)

    if not extra_properties:
        return

    if patterns:
        error = "Additional properties are not allowed (%s %s unexpected)"
        if len(extra_properties) == 1:
            verb = "was"
        else:
            verb = "were"
        yield jsonschema_exc.ValidationError(
            error
            % (", ".join(repr(extra) for extra in extra_properties), verb)
        )
    else:
        for prop in extra_properties:
            del param_value[prop]


def _validate_string_length(
    value,
    entity_name,
    mandatory=False,
    min_length=0,
    max_length=None,
    remove_whitespaces=False,
):
    """Check the length of specified string.

    :param value: the value of the string
    :param entity_name: the name of the string
    :mandatory: string is mandatory or not
    :param min_length: the min_length of the string
    :param max_length: the max_length of the string
    :param remove_whitespaces: True if trimming whitespaces is needed else
        False
    """
    if not mandatory and not value:
        return True

    if mandatory and not value:
        msg = _("The '%s' can not be None.") % entity_name
        raise webob.exc.HTTPBadRequest(explanation=msg)

    if remove_whitespaces:
        value = value.strip()

    utils.check_string_length(
        value, entity_name, min_length=min_length, max_length=max_length
    )


_FORMAT_CHECKER = jsonschema.FormatChecker()


@_FORMAT_CHECKER.checks("date-time")
def _validate_datetime_format(instance: object) -> bool:
    # format checks constrain to the relevant primitive type
    # https://github.com/OAI/OpenAPI-Specification/issues/3148
    if not isinstance(instance, str):
        return True
    try:
        timeutils.parse_isotime(instance)
    except ValueError:
        return False
    else:
        return True


@_FORMAT_CHECKER.checks("uuid")
def _validate_uuid_format(instance: object) -> bool:
    # format checks constrain to the relevant primitive type
    # https://github.com/OAI/OpenAPI-Specification/issues/3148
    if not isinstance(instance, str):
        return True

    return uuidutils.is_uuid_like(instance)


class _SchemaValidator:
    """A validator class.

    This class is changed from Draft202012Validator to validate minimum/maximum
    value of a string number(e.g. '10').

    In addition, FormatCheckers are added for checking data formats which are
    common in the Manila API.
    """

    validator = None
    validator_org = jsonschema.Draft202012Validator

    def __init__(
        self, schema, relax_additional_properties=False, is_body=True
    ):
        self.is_body = is_body
        validators = {
            "minimum": self._validate_minimum,
            "maximum": self._validate_maximum,
        }
        if relax_additional_properties:
            validators["additionalProperties"] = (
                _soft_validate_additional_properties
            )

        validator_cls = jsonschema.validators.extend(
            self.validator_org, validators
        )
        self.validator = validator_cls(schema, format_checker=_FORMAT_CHECKER)

    def validate(self, *args, **kwargs):
        try:
            self.validator.validate(*args, **kwargs)
        except jsonschema.ValidationError as ex:
            if len(ex.path) > 0:
                if self.is_body:
                    # NOTE: For consistency across OpenStack services, this
                    # error message has been written in a similar format as
                    # WSME errors.
                    detail = _(
                        "Invalid input for field/attribute %(path)s. "
                        "Value: %(value)s. %(message)s"
                    ) % {
                        "path": ex.path.pop(),
                        "value": ex.instance,
                        "message": ex.message,
                    }
                else:
                    # NOTE: We use 'ex.path.popleft()' instead of
                    # 'ex.path.pop()'. This is due to the structure of query
                    # parameters which is a dict with key as name and value is
                    # list. As such, the first item in the 'ex.path' is the key
                    # and second item is the index of list in the value. We
                    # need the key as the parameter name in the error message
                    # so we pop the first value out of 'ex.path'.
                    detail = _(
                        "Invalid input for query parameters %(path)s. "
                        "Value: %(value)s. %(message)s"
                    ) % {
                        "path": ex.path.popleft(),
                        "value": ex.instance,
                        "message": ex.message,
                    }
            else:
                detail = ex.message
            raise exception.SchemaValidationError(detail=detail)
        except TypeError as ex:
            # NOTE: If passing non string value to patternProperties parameter,
            # TypeError happens. Here is for catching the TypeError.
            detail = str(ex)
            raise exception.SchemaValidationError(detail=detail)

    def _number_from_str(self, param_value):
        try:
            value = int(param_value)
        except (ValueError, TypeError):
            try:
                value = float(param_value)
            except (ValueError, TypeError):
                return None
        return value

    def _validate_minimum(self, validator, minimum, param_value, schema):
        param_value = self._number_from_str(param_value)
        if param_value is None:
            return
        return self.validator_org.VALIDATORS["minimum"](
            validator, minimum, param_value, schema
        )

    def _validate_maximum(self, validator, maximum, param_value, schema):
        param_value = self._number_from_str(param_value)
        if param_value is None:
            return
        return self.validator_org.VALIDATORS["maximum"](
            validator, maximum, param_value, schema
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the unified limit backend driver in the
`keystone.unified_limit` namespace. Keystone only provides a `sql` driver, so
there's no reason to change this unless you are providing a custom entry point.
"""
    ),
)

caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt(
        """
Toggle for unified limit caching. This has no effect unless global caching is
enabled. In a typical deployment, there is no reason to disable this.
"""
    ),
)

cache_time = cfg.IntOpt(
    'cache_time',
    help=utils.fmt(
        """
Time to cache unified limit data, in seconds. This has no effect unless both
global caching and `[unified_limit] caching` are enabled.
"""
    ),
)

list_limit = cfg.IntOpt(
    'list_limit',
    help=utils.fmt(
        """
Maximum number of entities that will be returned in a unified limit
collection. This may be useful to tune if you have a large number of
unified limits in your deployment.
"""
    ),
)

enforcement_model = cfg.StrOpt(
    'enforcement_model',
    default='flat',
    choices=['flat', 'strict_two_level'],
    help=utils.fmt(
        """
The enforcement model to use when validating limits associated to projects.
Enforcement models will behave differently depending on the existing limits,
which may result in backwards incompatible changes if a model is switched in a
running deployment.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [driver, caching, cache_time, list_limit, enforcement_model]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any

from keystone.api.validation import parameter_types
from keystone.api.validation import response_types
from keystone.assignment.role_backends import resource_options as ro
from keystone.common import validation

_role_response_properties = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "description": validation.nullable(parameter_types.description),
            "domain_id": validation.nullable(parameter_types.domain_id),
            "id": parameter_types.id_string,
            "name": parameter_types.name,
            "options": ro.ROLE_OPTIONS_REGISTRY.json_schema,
            "links": response_types.links,
        },
        "additionalProperties": False,
    },
}

_trust_properties = {
    # NOTE(lbragstad): These are set as external_id_string because they have
    # the ability to be read as LDAP user identifiers, which could be something
    # other than uuid.
    "trustor_user_id": {
        "type": "string",
        "description": (
            "Represents the user who created the trust, and who's "
            "authorization is being delegated."
        ),
    },
    "trustee_user_id": {
        "type": "string",
        "description": "Represents the user who is capable of consuming "
        "the trust.",
    },
    "impersonation": {
        "type": "boolean",
        "description": (
            "If set to true, then the user attribute of tokens generated "
            "based on the trust will represent that of the trustor rather "
            "than the trustee, thus allowing the trustee to impersonate "
            "the trustor. If impersonation if set to false, then the token's "
            "user attribute will represent that of the trustee."
        ),
    },
    "project_id": {
        "type": ["string", "null"],
        "format": "uuid",
        "description": (
            "Identifies the project upon which the trustor is "
            "delegating authorization."
        ),
    },
    "remaining_uses": {
        "type": ["integer", "null"],
        "minimum": 1,
        "description": (
            "Specifies how many times the trust can be used to obtain "
            "a token. This value is decreased each time a token is issued "
            "through the trust. Once it reaches 0, no further tokens will "
            "be issued through the trust. The default value is null, "
            "meaning there is no limit on the number of tokens issued "
            "through the trust. If redelegation is enabled it must "
            "not be set."
        ),
    },
    "expires_at": {
        "type": ["null", "string"],
        "description": (
            "Specifies the expiration time of the trust. A trust may "
            "be revoked ahead of expiration. If the value represents "
            "a time in the past, the trust is deactivated. In the "
            "redelegation case it must not exceed the value of the "
            "corresponding expires_at field of the redelegated trust "
            "or it may be ommitted, then the expires_at value is copied "
            "from the redelegated trust."
        ),
    },
    "allow_redelegation": {
        "type": ["boolean", "null"],
        "description": (
            "If set to true then a trust between a trustor and any "
            "third-party user may be issued by the trustee just like a "
            "regular trust. If set to false, stops further redelegation. "
            "False by default."
        ),
    },
    "redelegation_count": {
        "type": ["integer", "null"],
        "minimum": 0,
        "description": (
            "Specifies the maximum remaining depth of the redelegated "
            "trust chain. Each subsequent trust has this field decremented "
            "by 1 automatically. The initial trustor issuing new trust "
            "that can be redelegated, must set allow_redelegation to true "
            "and may set redelegation_count to an integer value less than "
            "or equal to max_redelegation_count configuration parameter "
            "in order to limit the possible length of derivated trust chains. "
            "The trust issued by the trustor using a project-scoped token "
            "(not redelegating), in which allow_redelegation is set to true "
            "(the new trust is redelegatable), will be populated with the "
            "value specified in the max_redelegation_count configuration "
            "parameter if redelegation_count is not set or set to null. If "
            "allow_redelegation is set to false then redelegation_count "
            "will be set to 0 in the trust. If the trust is being issued by "
            "the trustee of a redelegatable trust-scoped token (redelegation "
            "case) then redelegation_count should not be set, as it will "
            "automatically be set to the value in the redelegatable "
            "trust-scoped token decremented by 1. Note, if the resulting "
            "value is 0, this means that the new trust will not be "
            "redelegatable, regardless of the value of allow_redelegation."
        ),
    },
    "redelegated_trust_id": {
        "type": ["string", "null"],
        "description": (
            "Returned with redelegated trust provides information "
            "about the predecessor in the trust chain.",
        ),
    },
}

trust_schema: dict[str, Any] = {
    "type": "object",
    "description": "A trust object.",
    "properties": {
        "deleted_at": {"type": ["string", "null"]},
        "id": {
            "type": "string",
            "readOnly": True,
            "description": "The ID of the trust.",
        },
        "links": response_types.links,
        "roles": _role_response_properties,
        "roles_links": response_types.links,
        **_trust_properties,
    },
    "additionalProperties": False,
}

trust_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "trustor_user_id": {
            "type": "string",
            "description": (
                "Represents the user who created the trust, and who's "
                "authorization is being delegated."
            ),
        },
        "trustee_user_id": {
            "type": "string",
            "description": "Represents the user who is capable of consuming "
            "the trust.",
        },
    },
    "additionalProperties": False,
}

trust_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "trusts": {
            "type": "array",
            "items": trust_schema,
            "description": "A list of trust objects.",
        },
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

trust_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {},
    "additionalProperties": False,
}

trust_response_body: dict[str, Any] = {
    "type": "object",
    "description": "A trust object",
    "properties": {"trust": trust_schema},
    "additionalProperties": False,
}

trust_create_request_body: dict[str, Any] = {
    "type": "object",
    "description": "A trust object",
    "properties": {
        "trust": {
            "type": "object",
            "properties": {
                **_trust_properties,
                "roles": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "id": parameter_types.id_string,
                            "name": parameter_types.name,
                        },
                        "minProperties": 1,
                        "maxProperties": 1,
                        "additionalProperties": False,
                    },
                },
            },
            "required": [
                "trustor_user_id",
                "trustee_user_id",
                "impersonation",
            ],
            "additionalProperties": True,
        }
    },
    "additionalProperties": False,
    "required": ["trust"],
}



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log
from pycadf import cadftaxonomy as taxonomy
from pycadf import reason
from pycadf import resource

from keystone.common import driver_hints
from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone import notifications

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs
_NOTIFY_OP = 'authenticate'
_NOTIFY_EVENT = f'{notifications.SERVICE}.{_NOTIFY_OP}'


def construct_method_map_from_config():
    """Determine authentication method types for deployment.

    :returns: a dictionary containing the methods and their indexes

    """
    method_map = {}
    method_index = 1
    for method in CONF.auth.methods:
        method_map[method_index] = method
        method_index = method_index * 2

    return method_map


def convert_method_list_to_integer(methods):
    """Convert the method type(s) to an integer.

    :param methods: a list of method names
    :returns: an integer representing the methods

    """
    method_map = construct_method_map_from_config()

    method_ints = []
    for method in methods:
        for k, v in method_map.items():
            if v == method:
                method_ints.append(k)
    return sum(method_ints)


def convert_integer_to_method_list(method_int):
    """Convert an integer to a list of methods.

    :param method_int: an integer representing methods
    :returns: a corresponding list of methods

    """
    # If the method_int is 0 then no methods were used so return an empty
    # method list
    if method_int == 0:
        return []

    method_map = construct_method_map_from_config()
    method_ints = sorted(method_map, reverse=True)

    methods = []
    for m_int in method_ints:
        # (lbragstad): By dividing the method_int by each key in the
        # method_map, we know if the division results in an integer of 1, that
        # key was used in the construction of the total sum of the method_int.
        # In that case, we should confirm the key value and store it so we can
        # look it up later. Then we should take the remainder of what is
        # confirmed and the method_int and continue the process. In the end, we
        # should have a list of integers that correspond to indexes in our
        # method_map and we can reinflate the methods that the original
        # method_int represents.
        result = int(method_int / m_int)
        if result == 1:
            methods.append(method_map[m_int])
            method_int = method_int - m_int

    return methods


class BaseUserInfo(provider_api.ProviderAPIMixin):
    @classmethod
    def create(cls, auth_payload, method_name):
        user_auth_info = cls()
        user_auth_info._validate_and_normalize_auth_data(auth_payload)
        user_auth_info.METHOD_NAME = method_name
        return user_auth_info

    def __init__(self):
        self.user_id = None
        self.user_ref = None
        self.METHOD_NAME = None

    def _assert_domain_is_enabled(self, domain_ref):
        try:
            PROVIDERS.resource_api.assert_domain_enabled(
                domain_id=domain_ref['id'], domain=domain_ref
            )
        except AssertionError as e:
            LOG.warning(e)
            raise exception.Unauthorized from e

    def _assert_user_is_enabled(self, user_ref):
        try:
            PROVIDERS.identity_api.assert_user_enabled(
                user_id=user_ref['id'], user=user_ref
            )
        except AssertionError as e:
            LOG.warning(e)
            raise exception.Unauthorized from e

    def _lookup_domain(self, domain_info):
        domain_id = domain_info.get('id')
        domain_name = domain_info.get('name')
        if not domain_id and not domain_name:
            raise exception.ValidationError(
                attribute='id or name', target='domain'
            )
        try:
            if domain_name:
                domain_ref = PROVIDERS.resource_api.get_domain_by_name(
                    domain_name
                )
            else:
                domain_ref = PROVIDERS.resource_api.get_domain(domain_id)
        except exception.DomainNotFound as e:
            LOG.warning(e)
            raise exception.Unauthorized(e)
        self._assert_domain_is_enabled(domain_ref)
        return domain_ref

    def _validate_and_normalize_auth_data(self, auth_payload):
        if 'user' not in auth_payload:
            raise exception.ValidationError(
                attribute='user', target=self.METHOD_NAME
            )
        user_info = auth_payload['user']
        user_id = user_info.get('id')
        user_name = user_info.get('name')
        domain_ref = {}
        if not user_id and not user_name:
            raise exception.ValidationError(
                attribute='id or name', target='user'
            )
        try:
            if user_name:
                if 'domain' not in user_info:
                    raise exception.ValidationError(
                        attribute='domain', target='user'
                    )
                domain_ref = self._lookup_domain(user_info['domain'])
                user_ref = PROVIDERS.identity_api.get_user_by_name(
                    user_name, domain_ref['id']
                )
            else:
                user_ref = PROVIDERS.identity_api.get_user(user_id)
                domain_ref = PROVIDERS.resource_api.get_domain(
                    user_ref['domain_id']
                )
                self._assert_domain_is_enabled(domain_ref)
        except exception.UserNotFound as e:
            LOG.warning(e)

            # We need to special case USER NOT FOUND here for CADF
            # notifications as the normal path for notification(s) come from
            # `identity_api.authenticate` and we are a bit before dropping into
            # that method.
            audit_reason = reason.Reason(str(e), str(e.code))
            audit_initiator = notifications.build_audit_initiator()
            # build an appropriate audit initiator with relevant information
            # for the failed request. This will catch invalid user_name and
            # invalid user_id.
            if user_name:
                audit_initiator.user_name = user_name
            else:
                audit_initiator.user_id = user_id
            audit_initiator.domain_id = domain_ref.get('id')
            audit_initiator.domain_name = domain_ref.get('name')
            notifications._send_audit_notification(
                action=_NOTIFY_OP,
                initiator=audit_initiator,
                outcome=taxonomy.OUTCOME_FAILURE,
                target=resource.Resource(typeURI=taxonomy.ACCOUNT_USER),
                event_type=_NOTIFY_EVENT,
                reason=audit_reason,
            )
            raise exception.Unauthorized(e)
        self._assert_user_is_enabled(user_ref)
        self.user_ref = user_ref
        self.user_id = user_ref['id']
        self.domain_id = domain_ref['id']


class UserAuthInfo(BaseUserInfo):
    def __init__(self):
        super().__init__()
        self.password = None

    def _validate_and_normalize_auth_data(self, auth_payload):
        super()._validate_and_normalize_auth_data(auth_payload)
        user_info = auth_payload['user']
        self.password = user_info.get('password')


class TOTPUserInfo(BaseUserInfo):
    def __init__(self):
        super().__init__()
        self.passcode = None

    def _validate_and_normalize_auth_data(self, auth_payload):
        super()._validate_and_normalize_auth_data(auth_payload)
        user_info = auth_payload['user']
        self.passcode = user_info.get('passcode')


class AppCredInfo(BaseUserInfo):
    def __init__(self):
        super().__init__()
        self.id = None
        self.secret = None

    def _validate_and_normalize_auth_data(self, auth_payload):
        app_cred_api = PROVIDERS.application_credential_api
        if auth_payload.get('id'):
            app_cred = app_cred_api.get_application_credential(
                auth_payload['id']
            )
            self.user_id = app_cred['user_id']
            if not auth_payload.get('user'):
                auth_payload['user'] = {}
                auth_payload['user']['id'] = self.user_id
            super()._validate_and_normalize_auth_data(auth_payload)
        elif auth_payload.get('name'):
            super()._validate_and_normalize_auth_data(auth_payload)
            hints = driver_hints.Hints()
            hints.add_filter('name', auth_payload['name'])
            app_cred = app_cred_api.list_application_credentials(
                self.user_id, hints
            )[0]
            auth_payload['id'] = app_cred['id']
        else:
            raise exception.ValidationError(
                attribute='id or name', target='application credential'
            )
        self.id = auth_payload['id']
        self.secret = auth_payload.get('secret')



#
# Copyright 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

from keystone.cmd import cli

# If ../../keystone/__init__.py exists, add ../../ to Python search path, so
# that it will override what happens to be installed in
# /usr/(local/)lib/python...
possible_topdir = os.path.normpath(
    os.path.join(os.path.abspath(__file__), os.pardir, os.pardir, os.pardir)
)
if os.path.exists(os.path.join(possible_topdir, 'keystone', '__init__.py')):
    sys.path.insert(0, possible_topdir)


# entry point.
def main():
    developer_config = os.path.join(possible_topdir, 'etc', 'keystone.conf')
    if not os.path.exists(developer_config):
        developer_config = None
    cli.main(argv=sys.argv, developer_config_file=developer_config)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

import keystone.conf
from keystone import exception

CONF = keystone.conf.CONF


class PolicyDriverBase(metaclass=abc.ABCMeta):
    def _get_list_limit(self):
        return CONF.policy.list_limit or CONF.list_limit

    @abc.abstractmethod
    def enforce(self, context, credentials, action, target):
        """Verify that a user is authorized to perform action.

        For more information on a full implementation of this see:
        `keystone.policy.backends.rules.Policy.enforce`
        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def create_policy(self, policy_id, policy):
        """Store a policy blob.

        :raises keystone.exception.Conflict: If a duplicate policy exists.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def list_policies(self):
        """List all policies."""
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def get_policy(self, policy_id):
        """Retrieve a specific policy blob.

        :raises keystone.exception.PolicyNotFound: If the policy doesn't exist.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def update_policy(self, policy_id, policy):
        """Update a policy blob.

        :raises keystone.exception.PolicyNotFound: If the policy doesn't exist.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def delete_policy(self, policy_id):
        """Remove a policy blob.

        :raises keystone.exception.PolicyNotFound: If the policy doesn't exist.

        """
        raise exception.NotImplemented()  # pragma: no cover



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file handles all flask-restful resources for /v3/role_inferences

import flask_restful

from keystone.api._shared import implied_roles as shared
from keystone.api import validation
from keystone.assignment import schema
from keystone.common import provider_api
from keystone.common import rbac_enforcer
from keystone.server import flask as ks_flask

ENFORCER = rbac_enforcer.RBACEnforcer
PROVIDERS = provider_api.ProviderAPIs


class RoleInferencesResource(flask_restful.Resource):
    @validation.request_body_schema(None)
    @validation.response_body_schema(
        schema.role_inferences_index_response_body
    )
    def get(self):
        """List role inference rules.

        GET/HEAD /v3/role_inferences
        """
        ENFORCER.enforce_call(action='identity:list_role_inference_rules')
        refs = PROVIDERS.role_api.list_role_inference_rules()
        role_dict = {
            role_ref['id']: role_ref
            for role_ref in PROVIDERS.role_api.list_roles()
        }

        rules = {}
        for ref in refs:
            implied_role_id = ref['implied_role_id']
            prior_role_id = ref['prior_role_id']
            implied = rules.get(prior_role_id, [])
            implied.append(
                shared.build_implied_role_response_data(
                    role_dict[implied_role_id]
                )
            )
            rules[prior_role_id] = implied

        inferences = []
        for prior_id, implied in rules.items():
            prior_response = shared.build_prior_role_response_data(
                prior_id, role_dict[prior_id]['name']
            )
            inferences.append(
                {'prior_role': prior_response, 'implies': implied}
            )
        results = {'role_inferences': inferences}
        return results


class RoleInferencesAPI(ks_flask.APIBase):
    _name = 'role_inferences'
    _import_name = __name__
    resources = []
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=RoleInferencesResource,
            url='/role_inferences',
            resource_kwargs={},
            rel='role_inferences',
        )
    ]


APIs = (RoleInferencesAPI,)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any


from keystone.api.validation import parameter_types
from keystone.api.validation import response_types
from keystone.assignment.schema import role_schema
from keystone.common import validation
from keystone.common.validation import parameter_types as old_parameter_types
from keystone.resource.backends import resource_options as ro

_name_properties = {
    'type': 'string',
    'description': 'The resource name.',
    'minLength': 1,
    'maxLength': 64,
    'pattern': r'[\S]+',
}

_project_tag_name_properties = {
    'type': 'string',
    'minLength': 1,
    'maxLength': 255,
    # NOTE(gagehugo) This pattern is for tags which follows the
    # guidelines as set by the API-WG, which matches anything that
    # does not contain a '/' or ','.
    # https://specs.openstack.org/openstack/api-wg/guidelines/tags.html
    'pattern': '^[^,/]*$',
}

_project_tags_list_properties = {
    'type': 'array',
    'items': _project_tag_name_properties,
    'required': [],
    'maxItems': 80,
    'uniqueItems': True,
}

_project_properties = {
    'description': validation.nullable(old_parameter_types.description),
    # NOTE(htruta): domain_id is nullable for projects acting as a domain.
    'domain_id': validation.nullable(old_parameter_types.id_string),
    'enabled': parameter_types.boolean,
    'is_domain': parameter_types.boolean,
    'parent_id': validation.nullable(old_parameter_types.id_string),
    'name': _name_properties,
    'tags': _project_tags_list_properties,
    'options': ro.PROJECT_OPTIONS_REGISTRY.json_schema,
}

# This is for updating a single project tag via the URL
project_tag_create = _project_tag_name_properties

# This is for updaing a project with a list of tags
project_tags_update = _project_tags_list_properties

project_create = {
    'type': 'object',
    'properties': _project_properties,
    # NOTE(lbragstad): A project name is the only parameter required for
    # project creation according to the Identity V3 API. We should think
    # about using the maxProperties validator here, and in update.
    'required': ['name'],
    'additionalProperties': True,
}

project_update = {
    'type': 'object',
    'properties': _project_properties,
    # NOTE(lbragstad): Make sure at least one property is being updated
    'minProperties': 1,
    'additionalProperties': True,
}

project_index_request_query = {
    'type': 'object',
    'properties': {
        "domain_id": parameter_types.domain_id,
        "enabled": parameter_types.boolean,
        "name": _name_properties,
        "parent_id": parameter_types.parent_id,
        "is_domain": parameter_types.boolean,
        "tags": parameter_types.tags,
        "tags-any": parameter_types.tags,
        "not-tags": parameter_types.tags,
        "not-tags-any": parameter_types.tags,
        "marker": {
            "type": "string",
            "description": "ID of the last fetched entry",
        },
        "limit": {"type": ["integer", "string"]},
    },
}

project_schema: dict[str, Any] = {
    "type": "object",
    "properties": {
        "id": {"type": "string", "readOnly": True},
        "links": response_types.resource_links,
        **_project_properties,
    },
    "additionalProperties": True,
}

project_get_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {"project": project_schema},
    "additionalProperties": False,
}

project_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "projects": {"type": "array", "items": project_schema},
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

project_create_request_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "project": {
            "type": "object",
            "properties": _project_properties,
            "required": ["name"],
        }
    },
    "additionalProperties": False,
}

project_create_response_body: dict[str, Any] = project_get_response_body

# Explicitly list attributes allowed to be updated.
# Since updating `domain_id` is since very long time marked deprecated do not
# even include it in the schema
project_update_request_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "project": {
            "type": "object",
            "properties": {
                'description': validation.nullable(
                    old_parameter_types.description
                ),
                'enabled': parameter_types.boolean,
                'name': _name_properties,
                'options': ro.PROJECT_OPTIONS_REGISTRY.json_schema,
                'tags': _project_tags_list_properties,
            },
        }
    },
    "additionalProperties": False,
}

project_update_response_body: dict[str, Any] = project_get_response_body

_domain_properties = {
    "description": validation.nullable(parameter_types.description),
    "enabled": {
        "description": "If set to true, domain is enabled. If set to false, domain is disabled.",
        **parameter_types.boolean,
    },
    "name": _name_properties,
    "options": ro.PROJECT_OPTIONS_REGISTRY.json_schema,
    "tags": project_tags_update,
}

domain_schema: dict[str, Any] = {
    "type": "object",
    "properties": {
        "id": {"type": "string", "readOnly": True},
        "links": response_types.resource_links,
        **_domain_properties,
    },
    "additionalProperties": False,
}

domain_get_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {"domain": domain_schema},
    "required": ["domain"],
    "additionalProperties": False,
}

domain_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "enabled": {
            "description": "If set to true, then only domains that are enabled will be returned, if set to false only that are disabled will be returned. Any value other than 0, including no value, will be interpreted as true.",
            **parameter_types.boolean,
        },
        "name": _name_properties,
        "marker": {
            "type": "string",
            "description": "ID of the last fetched entry",
        },
        "limit": {"type": ["integer", "string"]},
    },
    "additionalProperties": "False",
}

domain_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "domains": {"type": "array", "items": domain_schema},
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

domain_update_request_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "domain": {
            "type": "object",
            "properties": _domain_properties,
            "minProperties": 1,
        }
    },
    "required": ["domain"],
    "additionalProperties": False,
}

domain_update_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {"domain": domain_schema},
    "required": ["domain"],
    "additionalProperties": False,
}

domain_create_request_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "domain": {
            "type": "object",
            "properties": {
                "explicit_domain_id": {
                    "description": "The ID of the domain. A domain created this way will not use an auto-generated ID, but will use the ID passed in instead. Identifiers passed in this way must conform to the existing ID generation scheme: UUID4 without dashes.",
                    **parameter_types.domain_id,
                },
                **_domain_properties,
            },
            "required": ["name"],
        }
    },
    "required": ["domain"],
    "additionalProperties": False,
}

domain_create_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {"domain": domain_schema},
    "required": ["domain"],
    "additionalProperties": False,
}

tags_response_body: dict[str, Any] = response_types.tags

tags_update_request_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "tags": {"type": "array", "items": _project_tag_name_properties}
    },
    "additionalProperties": False,
}

# Response body of the `GET /[projects|domains]/{id}/[users|groups]/{id}/roles`
# API operations returning a list of roles
# Also used for the `GET /system/[users|groups]/{id}/roles` API operations
grants_get_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "links": response_types.links,
        "roles": {
            "type": "array",
            "items": {**role_schema, "additionalProperties": False},
        },
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures as db_fixtures
from oslotest import base as test_base

from keystone.common import sql
from keystone.identity.backends import sql as sql_backend
from keystone.tests.unit.identity.backends import test_base as id_test_base
from keystone.tests.unit.ksfixtures import database


class TestIdentityDriver(
    db_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
    id_test_base.IdentityDriverTests,
):
    expected_is_domain_aware = True
    expected_default_assignment_driver = 'sql'
    expected_is_sql = True
    expected_generates_uuids = True

    def setUp(self):
        super().setUp()
        self.engine = enginefacade.writer.get_engine()
        self.sessionmaker = enginefacade.writer.get_sessionmaker()

        # Set keystone's connection URL to be the test engine's url. Close
        # sqlite FK to avoid conflicting with sql upgrade test.
        database.initialize_sql_session(
            self.engine.url, enforce_sqlite_fks=False
        )

        # Override keystone's context manager to be oslo.db's global context
        # manager.
        sql.core._TESTING_USE_GLOBAL_CONTEXT_MANAGER = True
        self.addCleanup(
            setattr, sql.core, '_TESTING_USE_GLOBAL_CONTEXT_MANAGER', False
        )
        self.addCleanup(sql.cleanup)

        database._load_sqlalchemy_models()
        sql.ModelBase.metadata.create_all(bind=self.engine)

        self.driver = sql_backend.Identity()


class MySQLOpportunisticIdentityDriverTestCase(TestIdentityDriver):
    FIXTURE = db_fixtures.MySQLOpportunisticFixture


class PostgreSQLOpportunisticIdentityDriverTestCase(TestIdentityDriver):
    FIXTURE = db_fixtures.PostgresqlOpportunisticFixture



# Copyright 2018 Huawei
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any

from keystone.api.validation import parameter_types
from keystone.api.validation import response_types
from keystone.common import validation

_limit_integer_type = {
    "type": "integer",
    "minimum": -1,
    "maximum": 0x7FFFFFFF,  # The maximum value a signed INT may have
}

_registered_limit_properties = {
    "service_id": {
        "type": "string",
        "format": "uuid",
        "description": (
            "The UUID of the service to which the registered limit belongs."
        ),
    },
    "region_id": {
        "description": (
            "The ID of the region that contains the service endpoint."
        ),
        **parameter_types.region_id,
    },
    "resource_name": parameter_types.name,
    "default_limit": {
        "description": "The default limit for the registered limit.",
        **_limit_integer_type,
    },
    "description": validation.nullable(parameter_types.description),
}

# Common schema of `Registered Limit` resource
registered_limit_schema: dict[str, Any] = {
    "type": "object",
    "description": "A registered limit object.",
    "properties": {
        "id": {
            "type": "string",
            "format": "uuid",
            "description": "The registered limit ID.",
            "readOnly": True,
        },
        "links": response_types.resource_links,
        **_registered_limit_properties,
    },
    "additionalProperties": False,
}

# Response body of API operations returning a single registered limit
# `GET /registered_limits/{registered_limit_id}` and
# `PATCH /registered_limits/{registered_limit_id}`
registered_limit_show_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {"registered_limit": registered_limit_schema},
    "additionalProperties": False,
}

# Query parameters of the `GET /registered_limits` API operation
# returning a list of registered limits
registered_limits_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "service_id": {
            "type": "string",
            "format": "uuid",
            "description": ("Filters the response by a service ID."),
        },
        "region_id": {
            "description": ("Filters the response by a region ID."),
            **parameter_types.region_id,
        },
        "resource_name": {
            "description": {
                "Filters the response by a specified resource name."
            },
            **parameter_types.name,
        },
    },
    "additionalProperties": False,
}

# Response body of the `GET /registered_limits` API operation
# returning a list of registered limits
registered_limits_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "links": response_types.links,
        "registered_limits": {
            "type": "array",
            "items": registered_limit_schema,
            "description": "A list of registered limit objects.",
        },
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

# Individual properties for creating a new registered limit
_registered_limit_create = {
    "type": "object",
    "properties": _registered_limit_properties,
    "additionalProperties": False,
    "required": ["service_id", "resource_name", "default_limit"],
}

# Request body of the `POST /registered_limits` API operation
registered_limits_create_request_body = {
    "type": "object",
    "properties": {
        "registered_limits": {
            "type": "array",
            "items": _registered_limit_create,
            "minItems": 1,
            "description": "A list of registered limit objects.",
        }
    },
    "required": ["registered_limits"],
    "additionalProperties": False,
}

# Response body of the `POST /registered_limits` API operation
registered_limits_create_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "registered_limits": {
            "type": "array",
            "items": registered_limit_schema,
            "description": "A list of registered limit objects.",
        },
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

# Request body of the `PATCH /registered_limits/{registered_limit_id}`
# operation
registered_limit_update_request_body = {
    "type": "object",
    "properties": {
        "registered_limit": {
            "type": "object",
            "description": "Updates to make to a registered limit.",
            "properties": {
                "service_id": {
                    "type": "string",
                    "format": "uuid",
                    "description": (
                        "The UUID of the service to which "
                        "the registered limit belongs."
                    ),
                },
                "region_id": parameter_types.region_id,
                "resource_name": parameter_types.name,
                "default_limit": _limit_integer_type,
                "description": validation.nullable(
                    parameter_types.description
                ),
            },
            "additionalProperties": False,
        }
    },
    "additionalProperties": False,
    "required": ["registered_limit"],
}

# Individual properties of the `Limit`
_limit_properties = {
    "resource_name": parameter_types.name,
    "region_id": {
        "description": (
            "The ID of the region that contains the service endpoint."
        ),
        **parameter_types.region_id,
    },
    "service_id": {
        "type": "string",
        "format": "uuid",
        "description": "The UUID of the service to which the limit belongs.",
    },
    "resource_limit": {
        "description": "The override limit.",
        **_limit_integer_type,
    },
    "description": validation.nullable(parameter_types.description),
}

# Common schema of `Limit` resource
limit_schema: dict[str, Any] = {
    "type": "object",
    "description": "A limit object.",
    "properties": {
        "id": {
            "type": "string",
            "format": "uuid",
            "description": "The limit ID.",
            "readOnly": True,
        },
        "project_id": validation.nullable(parameter_types.project_id),
        "domain_id": validation.nullable(parameter_types.domain_id),
        "links": response_types.resource_links,
        **_limit_properties,
    },
    "additionalProperties": False,
}

# Response body of API operations returning a single limit
# `GET /limits/{limit_id}` and `PATCH /limits/{limit_id}`
limit_show_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {"limit": limit_schema},
    "additionalProperties": False,
}

# Query parameters of the `GET /limits` API operation
# returning a list of limits
limits_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "service_id": {
            "type": "string",
            "format": "uuid",
            "description": "Filters the response by a service ID.",
        },
        "region_id": {
            "description": "Filters the response by a region ID.",
            **parameter_types.region_id,
        },
        "resource_name": {
            "description": (
                "Filters the response by a specified resource name."
            ),
            **parameter_types.name,
        },
        "project_id": {
            "description": "Filters the response by a project ID.",
            **parameter_types.project_id,
        },
        "domain_id": {
            "description": "Filters the response by a domain ID.",
            **parameter_types.domain_id,
        },
    },
    "additionalProperties": False,
}

# Response body of the `GET /limits` API operation
# returning a list of limits
limits_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "links": response_types.links,
        "limits": {
            "type": "array",
            "items": limit_schema,
            "description": "A list of limit objects.",
        },
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

# Response body of the `GET /limits/model` API operation
# returning an enforcement model
limit_model_show_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "model": {
            "type": "object",
            "description": (
                "A model object describing the configured enforcement model "
                "used by the deployment."
            ),
            "properties": {
                "description": {
                    "type": "string",
                    "description": (
                        "A short description of the enforcement model used."
                    ),
                },
                "name": {
                    **parameter_types.name,
                    "description": "The name of the enforcement model.",
                },
            },
        },
        "additionalProperties": False,
    },
    "additionalProperties": False,
}

# Individual properties for creating a new limit
_limit_create = {
    "type": "object",
    "properties": {
        "project_id": validation.nullable(parameter_types.project_id),
        "domain_id": validation.nullable(parameter_types.domain_id),
        **_limit_properties,
    },
    "required": ["service_id", "resource_name", "resource_limit"],
    "oneOf": [
        {
            "required": [
                "service_id",
                "resource_name",
                "resource_limit",
                "domain_id",
            ]
        },
        {
            "required": [
                "service_id",
                "resource_name",
                "resource_limit",
                "project_id",
            ]
        },
    ],
    "additionalProperties": False,
}

# Request body of the `POST /limits` API operation
limits_create_request_body = {
    "type": "object",
    "properties": {
        "limits": {
            "type": "array",
            "items": _limit_create,
            "minItems": 1,
            "description": "A list of limit objects.",
        }
    },
    "required": ["limits"],
    "additionalProperties": False,
}

# Response body of the `POST /limits` API operation
limits_create_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "limits": {
            "type": "array",
            "items": limit_schema,
            "description": "A list of limit objects.",
        }
    },
    "additionalProperties": False,
}

# Request body of the `PATCH /limits/{limit_id}` operation
limit_update_request_body = {
    "type": "object",
    "properties": {
        "limit": {
            "type": "object",
            "description": "Updates to make to a limit.",
            "properties": {
                "resource_limit": {
                    "description": "The override limit.",
                    **_limit_integer_type,
                },
                "description": validation.nullable(
                    parameter_types.description
                ),
            },
            "additionalProperties": False,
        }
    },
    "additionalProperties": False,
    "required": ["limit"],
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock
import uuid

from keystone.assignment.core import Manager as AssignmentApi
from keystone.auth.plugins import mapped
from keystone.exception import ProjectNotFound
from keystone.resource.core import Manager as ResourceApi
from keystone.tests import unit


class TestMappedPlugin(unit.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        super().setUp()
        self.resource_api_mock = mock.Mock(spec=ResourceApi)
        self.assignment_api_mock = mock.Mock(spec=AssignmentApi)
        self.domain_uuid_mock = uuid.uuid4().hex
        self.domain_mock = {'id': self.domain_uuid_mock}
        self.idp_domain_uuid_mock = uuid.uuid4().hex
        self.member_role_id = uuid.uuid4().hex
        self.member_role_name = "member"
        self.existing_roles = {
            self.member_role_name: {'id': self.member_role_id}
        }
        self.shadow_project_mock = {
            'name': "test-project",
            'roles': [{'name': self.member_role_name}],
        }
        self.shadow_project_in_domain_mock = {
            'name': "test-project-in-domain",
            'domain': self.domain_mock,
            'roles': [{'name': self.member_role_name}],
        }
        self.shadow_projects_mock = [
            self.shadow_project_mock,
            self.shadow_project_in_domain_mock,
        ]
        self.user_mock = {'id': uuid.uuid4().hex, 'name': "test-user"}

    def test_configure_project_domain_no_project_domain(self):
        mapped.configure_project_domain(
            self.shadow_project_mock,
            self.idp_domain_uuid_mock,
            self.resource_api_mock,
        )
        self.assertIn("domain", self.shadow_project_mock)
        self.assertEqual(
            self.idp_domain_uuid_mock, self.shadow_project_mock['domain']['id']
        )

    def test_configure_project_domain_with_domain_id(self):
        self.shadow_project_mock['domain'] = self.domain_mock
        mapped.configure_project_domain(
            self.shadow_project_mock,
            self.idp_domain_uuid_mock,
            self.resource_api_mock,
        )
        self.assertIn("domain", self.shadow_project_mock)
        self.assertEqual(
            self.domain_uuid_mock, self.shadow_project_mock['domain']['id']
        )

    def test_configure_project_domain_with_domain_name(self):
        domain_name = "test-domain"
        self.shadow_project_mock['domain'] = {'name': domain_name}
        self.resource_api_mock.get_domain_by_name.return_value = (
            self.domain_mock
        )
        mapped.configure_project_domain(
            self.shadow_project_mock,
            self.idp_domain_uuid_mock,
            self.resource_api_mock,
        )
        self.assertIn("domain", self.shadow_project_mock)
        self.assertEqual(
            self.domain_uuid_mock, self.shadow_project_mock['domain']['id']
        )
        self.resource_api_mock.get_domain_by_name.assert_called_with(
            domain_name
        )

    def test_handle_projects_from_mapping_project_exists(self):
        project_mock_1 = self.create_project_mock_for_shadow_project(
            self.shadow_project_mock
        )
        project_mock_2 = self.create_project_mock_for_shadow_project(
            self.shadow_project_in_domain_mock
        )
        self.resource_api_mock.get_project_by_name.side_effect = [
            project_mock_1,
            project_mock_2,
        ]
        mapped.handle_projects_from_mapping(
            self.shadow_projects_mock,
            self.idp_domain_uuid_mock,
            self.existing_roles,
            self.user_mock,
            self.assignment_api_mock,
            self.resource_api_mock,
        )
        self.resource_api_mock.get_project_by_name.assert_has_calls(
            [
                mock.call(
                    self.shadow_project_in_domain_mock['name'],
                    self.shadow_project_in_domain_mock['domain']['id'],
                ),
                mock.call(
                    self.shadow_project_mock['name'], self.idp_domain_uuid_mock
                ),
            ],
            any_order=True,
        )
        self.assignment_api_mock.create_grant.assert_has_calls(
            [
                mock.call(
                    self.member_role_id,
                    user_id=self.user_mock['id'],
                    project_id=project_mock_1['id'],
                ),
                mock.call(
                    self.member_role_id,
                    user_id=self.user_mock['id'],
                    project_id=project_mock_2['id'],
                ),
            ]
        )

    @mock.patch("uuid.UUID.hex", new_callable=mock.PropertyMock)
    def test_handle_projects_from_mapping_create_projects(self, uuid_mock):
        uuid_mock.return_value = "uuid"
        project_mock_1 = self.create_project_mock_for_shadow_project(
            self.shadow_project_mock
        )
        project_mock_2 = self.create_project_mock_for_shadow_project(
            self.shadow_project_in_domain_mock
        )
        self.resource_api_mock.get_project_by_name.side_effect = [
            ProjectNotFound(project_id=project_mock_1['name']),
            ProjectNotFound(project_id=project_mock_2['name']),
        ]
        self.resource_api_mock.create_project.side_effect = [
            project_mock_1,
            project_mock_2,
        ]
        mapped.handle_projects_from_mapping(
            self.shadow_projects_mock,
            self.idp_domain_uuid_mock,
            self.existing_roles,
            self.user_mock,
            self.assignment_api_mock,
            self.resource_api_mock,
        )
        self.resource_api_mock.get_project_by_name.assert_has_calls(
            [
                mock.call(
                    self.shadow_project_in_domain_mock['name'],
                    self.shadow_project_in_domain_mock['domain']['id'],
                ),
                mock.call(
                    self.shadow_project_mock['name'], self.idp_domain_uuid_mock
                ),
            ],
            any_order=True,
        )
        expected_project_ref1 = {
            'id': "uuid",
            'name': self.shadow_project_mock['name'],
            'domain_id': self.idp_domain_uuid_mock,
        }
        expected_project_ref2 = {
            'id': "uuid",
            'name': self.shadow_project_in_domain_mock['name'],
            'domain_id': self.shadow_project_in_domain_mock['domain']['id'],
        }
        self.resource_api_mock.create_project.assert_has_calls(
            [
                mock.call(expected_project_ref1['id'], expected_project_ref1),
                mock.call(expected_project_ref2['id'], expected_project_ref2),
            ]
        )
        self.assignment_api_mock.create_grant.assert_has_calls(
            [
                mock.call(
                    self.member_role_id,
                    user_id=self.user_mock['id'],
                    project_id=project_mock_1['id'],
                ),
                mock.call(
                    self.member_role_id,
                    user_id=self.user_mock['id'],
                    project_id=project_mock_2['id'],
                ),
            ]
        )

    def create_project_mock_for_shadow_project(self, shadow_project):
        project = shadow_project.copy()
        project['id'] = uuid.uuid4().hex
        return project



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any

from keystone.api.validation import parameter_types
from keystone.api.validation import response_types
from keystone.assignment.role_backends import resource_options as ro
from keystone.common import validation

# Schema for Identity v3 API

_role_properties: dict[str, Any] = {
    "name": parameter_types.name,
    "description": validation.nullable(parameter_types.description),
    "domain_id": validation.nullable(parameter_types.domain_id),
    "options": ro.ROLE_OPTIONS_REGISTRY.json_schema,
}
# NOTE(0weng): Multiple response body examples in the docs are
# incorrectly missing the `options` field.

# Common schema of `Role` resource
role_schema: dict[str, Any] = {
    "type": "object",
    "description": "A role object.",
    "properties": {
        "id": {
            "type": "string",
            "format": "uuid",
            "description": "The role ID.",
            "readOnly": True,
        },
        "links": response_types.resource_links,
        **_role_properties,
    },
    "additionalProperties": True,
}

# Response body of API operations returning a single role
# `GET /roles/{role_id}`, `POST /roles`, and `PATCH /roles/{role_id}`
role_show_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {"role": role_schema},
    "additionalProperties": False,
}

# Query parameters of the `GET /roles` API operation
# returning a list of roles
roles_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "name": parameter_types.name,
        "domain_id": parameter_types.domain_id,
    },
    "additionalProperties": False,
}

# Response body of the `GET /roles` API operation
# returning a list of roles
roles_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "links": response_types.links,
        "roles": {
            "type": "array",
            "items": role_schema,
            "description": "A list of role objects.",
        },
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

# Request body of the `POST /roles` API operation
role_create_request_body = {
    "type": "object",
    "properties": {
        "role": {
            "type": "object",
            "properties": _role_properties,
            "description": "A role object.",
            "required": ["name"],
            "additionalProperties": True,
        }
    },
    "required": ["role"],
    "additionalProperties": False,
}

# FIXME(0weng): There's no error if additional properties are added
# at the top level, e.g. POST/PATCH with this body:
# {"role": {"some_key":"some_value"}, "no_error": "no_error_here"}
# Is this intended, or should it be disallowed by the schema (as it is here)?
# 400 errors do occur if the "role" property is missing
# (the error message is that '{}' is not enough properties,
# so I imagine extra properties are removed)
# or no properties are provided.

# Request body of the `PATCH /roles/{role_id}` operation
role_update_request_body = {
    "type": "object",
    "properties": {
        "role": {
            "type": "object",
            "properties": _role_properties,
            "description": "A role object.",
            "minProperties": 1,
            "additionalProperties": True,
        }
    },
    "required": ["role"],
    "additionalProperties": False,
}

# Individual properties of a returned prior/implied role
_implied_role_properties: dict[str, Any] = {
    "id": {
        "type": "string",
        "format": "uuid",
        "description": "The role ID.",
        "readOnly": True,
    },
    "links": response_types.resource_links,
    "name": parameter_types.name,
}

# Common schema of prior role
prior_role_schema: dict[str, Any] = {
    "type": "object",
    "description": "A prior role object.",
    "properties": _implied_role_properties,
    "additionalProperties": False,
}

# Common schema of implied role
implied_role_schema: dict[str, Any] = {
    "type": "object",
    "description": "An implied role object.",
    "properties": _implied_role_properties,
    "additionalProperties": False,
}

# Response body of API operations returning a single implied role
# `GET /v3/roles/{prior_role_id}/implies/{implies_role_id}`
# and `PUT /v3/roles/{prior_role_id}/implies/{implies_role_id}`
implied_role_show_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "role_inference": {
            "type": "object",
            "description": (
                "Role inference object that contains "
                "prior_role object and implies object."
            ),
            "properties": {
                "prior_role": prior_role_schema,
                "implies": implied_role_schema,
            },
            "additionalProperties": False,
        },
        "links": response_types.links,
    },
    "additionalProperties": False,
}

# Response body of the `GET /v3/roles/{prior_role_id}/implies` API operation
# returning a single role inference
role_inference_show_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "role_inference": {
            "type": "object",
            "description": "A role inference object.",
            "properties": {
                "prior_role": prior_role_schema,
                "implies": {
                    "type": "array",
                    "items": implied_role_schema,
                    "description": "A list of implied role objects.",
                },
            },
            "additionalProperties": False,
        },
        "links": response_types.links,
    },
    "additionalProperties": False,
}

# Response body of the `GET /v3/role_inferences` API operation
# returning a list of role inferences
role_inferences_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "role_inferences": {
            "type": "array",
            "description": "A list of role inference objects.",
            "items": {
                "type": "object",
                "properties": {
                    # NOTE(0weng): The example in the docs incorrectly
                    # includes the `description` field in the output.
                    "prior_role": prior_role_schema,
                    "implies": {
                        "type": "array",
                        "items": implied_role_schema,
                        "description": "A list of implied role objects.",
                    },
                },
                "additionalProperties": False,
            },
            "additionalProperties": False,
        },
        "links": response_types.links,
    },
    "additionalProperties": False,
}

# Common schema of `Role Assignment` resource
role_assignment_schema: dict[str, Any] = {
    "type": "object",
    "description": "A role assignment object.",
    "properties": {
        "links": {
            "type": "object",
            "properties": {
                "assignment": {"type": "string", "format": "uri"},
                "membership": {"type": "string", "format": "uri"},
                "prior_role": {"type": "string", "format": "uri"},
            },
            "required": ["assignment"],
            "additionalProperties": False,
        },
        "role": {
            "type": "object",
            "properties": {
                "domain": {
                    "type": "object",
                    "properties": {
                        "id": parameter_types.id_string,
                        "name": parameter_types.name,
                    },
                    "required": ["id"],
                    "additionalProperties": False,
                },
                "id": parameter_types.id_string,
                "name": parameter_types.name,
            },
            "required": ["id"],
            "additionalProperties": False,
        },
        "scope": {
            "properties": {
                "project": {
                    "type": "object",
                    "properties": {
                        "domain": {
                            "type": "object",
                            "properties": {
                                "id": parameter_types.id_string,
                                "name": parameter_types.name,
                            },
                            "required": ["id"],
                            "additionalProperties": False,
                        },
                        "id": parameter_types.id_string,
                        "name": parameter_types.name,
                    },
                    "required": ["id"],
                    "additionalProperties": False,
                },
                "OS-INHERIT:inherited_to": {"const": "projects"},
                "domain": {
                    "type": "object",
                    "properties": {
                        "id": parameter_types.id_string,
                        "name": parameter_types.name,
                    },
                    "required": ["id"],
                },
                "system": {
                    "type": "object",
                    "properties": {"all": {"const": True}},
                    "required": ["all"],
                },
            },
            "oneOf": [
                {"required": ["project"]},
                {"required": ["domain"]},
                {"required": ["system"]},
            ],
            "additionalProperties": False,
        },
        "user": {
            "type": "object",
            "properties": {
                "domain": {
                    "type": "object",
                    "properties": {
                        "id": parameter_types.id_string,
                        "name": parameter_types.name,
                    },
                    "required": ["id"],
                    "additionalProperties": False,
                },
                "id": parameter_types.id_string,
                "name": parameter_types.name,
            },
            "required": ["id"],
            "additionalProperties": False,
        },
        "group": {
            "type": "object",
            "properties": {
                "domain": {
                    "type": "object",
                    "properties": {
                        "id": parameter_types.id_string,
                        "name": parameter_types.name,
                    },
                    "required": ["id"],
                    "additionalProperties": False,
                },
                "id": parameter_types.id_string,
                "name": parameter_types.name,
            },
            "required": ["id"],
            "additionalProperties": False,
        },
    },
    "required": ["role", "links", "scope"],
    "oneOf": [{"required": ["user"]}, {"required": ["group"]}],
    "additionalProperties": False,
}

# Query parameters of the `GET /role_assignments` API operation
# returning a list of role assignments
role_assignments_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        # NOTE(0weng): `effective`, `include_names`, and `include_subtree`
        # are all key-only query parameters and will interpret the parameter
        # as true for any value except 0 and for no value at all.
        "effective": {},
        "include_names": {},
        "include_subtree": {},
        "group.id": parameter_types.id_string,
        "role.id": parameter_types.id_string,
        # NOTE(0weng): Similar to the previous note, `scope.system`
        # accepts anything, but it will only be interpreted as true if
        # its value is "all"; otherwise it's ignored.
        "scope.system": {},
        "scope.domain.id": parameter_types.domain_id,
        "scope.project.id": parameter_types.project_id,
        "user.id": parameter_types.id_string,
        # NOTE(0weng): `scope.OS-INHERIT:inherited_to` accepts anything,
        # but it will only show inherited role assignments if its value is
        # `projects`; otherwise, only non-inherited role assignments are shown.
        "scope.OS-INHERIT:inherited_to": {},
    },
    "dependentRequired": {"include_subtree": ["scope.project.id"]},
    # NOTE(0weng): {"not": {"required": "some_property"}} indicates that
    # the property is not allowed to be present, not that it's optional.
    "oneOf": [
        {"required": ["scope.project.id"]},
        {"required": ["scope.domain.id"]},
        {"required": ["scope.system"]},
        {
            "allOf": [
                {"not": {"required": ["scope.project.id"]}},
                {"not": {"required": ["scope.domain.id"]}},
                {"not": {"required": ["scope.system"]}},
            ]
        },
    ],
    "oneOf": [
        {"required": ["user.id"]},
        {"required": ["group.id"]},
        {
            "allOf": [
                {"not": {"required": ["user.id"]}},
                {"not": {"required": ["group.id"]}},
            ]
        },
    ],
    "not": {
        "required": [
            "effective",
            "scope.domain.id",
            "scope.OS-INHERIT:inherited_to",
        ]
    },
    "not": {"required": ["effective", "group"]},
    "additionalProperties": False,
}

# Response body of the `GET /role_assignments` API operation
# returning a list of role assignments
role_assignments_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "links": response_types.links,
        "role_assignments": {
            "type": "array",
            "items": role_assignment_schema,
            "description": "A list of role assignment objects.",
        },
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import flask
from oslo_log import log

from keystone.auth.plugins import base
from keystone.auth.plugins import mapped
from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.i18n import _

LOG = log.getLogger(__name__)

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class Token(base.AuthMethodHandler):
    def _get_token_ref(self, auth_payload):
        token_id = auth_payload['id']
        return PROVIDERS.token_provider_api.validate_token(token_id)

    def authenticate(self, auth_payload):
        if 'id' not in auth_payload:
            raise exception.ValidationError(attribute='id', target='token')
        token = self._get_token_ref(auth_payload)
        if token.is_federated and PROVIDERS.federation_api:
            response_data = mapped.handle_scoped_token(
                token, PROVIDERS.federation_api, PROVIDERS.identity_api
            )
        else:
            response_data = token_authenticate(token)

        # NOTE(notmorgan): The Token auth method is *very* special and sets the
        # previous values to the method_names. This is because it can be used
        # for re-scoping and we want to maintain the values. Most
        # AuthMethodHandlers do no such thing and this is not required.
        response_data.setdefault('method_names', []).extend(token.methods)

        return base.AuthHandlerResponse(
            status=True, response_body=None, response_data=response_data
        )


def token_authenticate(token):
    response_data = {}
    try:
        # Do not allow tokens used for delegation to
        # create another token, or perform any changes of
        # state in Keystone. To do so is to invite elevation of
        # privilege attacks

        json_body = flask.request.get_json(silent=True, force=True) or {}
        project_scoped = 'project' in json_body['auth'].get('scope', {})
        domain_scoped = 'domain' in json_body['auth'].get('scope', {})

        if token.oauth_scoped:
            raise exception.ForbiddenAction(
                action=_(
                    'Using OAuth-scoped token to create another token. '
                    'Create a new OAuth-scoped token instead'
                )
            )
        elif token.trust_scoped:
            raise exception.ForbiddenAction(
                action=_(
                    'Using trust-scoped token to create another token. '
                    'Create a new trust-scoped token instead'
                )
            )
        elif token.system_scoped and (project_scoped or domain_scoped):
            raise exception.ForbiddenAction(
                action=_(
                    'Using a system-scoped token to create a project-scoped '
                    'or domain-scoped token is not allowed.'
                )
            )

        if not CONF.token.allow_rescope_scoped_token:
            # Do not allow conversion from scoped tokens.
            if token.project_scoped or token.domain_scoped:
                raise exception.ForbiddenAction(
                    action=_('rescope a scoped token')
                )

        # New tokens maintain the audit_id of the original token in the
        # chain (if possible) as the second element in the audit data
        # structure. Look for the last element in the audit data structure
        # which will be either the audit_id of the token (in the case of
        # a token that has not been rescoped) or the audit_chain id (in
        # the case of a token that has been rescoped).
        try:
            token_audit_id = token.parent_audit_id or token.audit_id
        except IndexError:
            # NOTE(morganfainberg): In the case this is a token that was
            # issued prior to audit id existing, the chain is not tracked.
            token_audit_id = None

        # To prevent users from never having to re-authenticate, the original
        # token expiration time is maintained in the new token. Not doing this
        # would make it possible for a user to continuously bump token
        # expiration through token rescoping without proving their identity.
        response_data.setdefault('expires_at', token.expires_at)
        response_data['audit_id'] = token_audit_id
        response_data.setdefault('user_id', token.user_id)

        return response_data

    except AssertionError as e:
        LOG.error(e)
        raise exception.Unauthorized(e)
